/*
    Freud - a standalone C++ port of the Emacs Psychiatrist
    Copyright (C) 2002  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    You can contact me via triska@gmx.at
*/


#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include <vector>
#include <algorithm>

#include <ctype.h>
#include <stdlib.h>

#include "synonymlist.h"
#include "wordtypes.h"

using namespace std;


WordTypeDeterminer wordtypes;

// Remember here the latest verb, object, what the patient feared etc.
map<string, string> recententities;

map<string, string> meanings;
map<string, SynonymList> synonyms;
map<string, string> replist;
map<string, string> corrections;
map<string, string> expansions;

// separate words of the user's input
vector<string> sent;
typedef vector<string>::iterator sent_it;
typedef vector<string>::reverse_iterator sent_rev_it;

extern void init_corrections();
extern void init_meanings();
extern void init_synonyms();
extern void init_wordtypes();

unsigned int linecount = 0;
bool printspace = false;
bool printupcase = true;


void printsymbol(const string &s)
{
	string oneword = s;
	if (oneword[0] == '/') {
		// print a global variable
		string varname = string(oneword, 1, oneword.size());
		printsymbol(recententities[varname]);
	} else if (oneword[0] == '$') {
		// print a synonym from the list
		string synclass = string(oneword, 1, oneword.size());
		// cout << "Synclass: " << synclass << endl;
		if (synonyms.find(synclass) != synonyms.end()) {
			for_each_word(synonyms[synclass].current(),printsymbol);
		} else {
			cout << "ATTENTION: class " << synclass << " does not exist!" << endl;
		}
	} else {
		if (oneword == "i") oneword == "I";
		if (!ispunct(oneword[0]) && !isspace(oneword[0])) {
			if (printspace) cout << ' ';
		}
		if (printupcase) oneword[0] = toupper(oneword[0]);
		// The Emacs doctor puts another space after
		// punctuation. We don't.

		cout << oneword;
		printspace = true;
		switch(oneword[oneword.size() - 1]) {
		case '.': case '?': case '!':
			printupcase = true; break;
		default:
			printupcase = false;
		}
	}
}

void fix_input_2();

void printline(const string &s)
{
	printupcase = true;
	printspace = false;
	fix_input_2();

	// Save the corrected and perspective-swapped input for possible reply
	recententities["sent"] = "";
	for (sent_it sit = sent.begin(); sit != sent.end(); sit++) {
		if (sit != sent.begin()) recententities["sent"] += ' ';
		recententities["sent"] += *sit;
	}

	for_each_word(s, printsymbol);
	cout << endl;
}

string trim(const string &s)
{
	// Remove whitespace from begin and end of the given string
	unsigned int startpos = 0, len = s.size() - 1;
	while (true) {
		if (startpos >= (s.size() - 1)) break;
		if (!isspace(s[startpos]) && !isspace(s[len])) break;
		if (isspace(s[startpos])) startpos++;
		if (isspace(s[len])) len--;
	}
	return s.substr(startpos, ++len);
}

bool sentcontains(const string s)
{
	return find(sent.begin(), sent.end(), s) != sent.end();
}

void push_to_sent(const string &s)
{
	sent.push_back(s);
}

// WordCorrector is a helper classes so that we can use for_each_word() on a
// string of words separated by whitespace.

class WordCorrector
{
public:
	WordCorrector() { }
	string corrected() const { return corr; }
	void operator() (string s) {
		// We could use the expansions-map also to expand (correct)
		// contractions (e.g., "they'll"). Emacs doctor does not.
		if (corrections.find(s) != corrections.end()) {
			s = expansions[corrections[s]];
		} else if (replist.find(s) != replist.end()) {
			s = replist[s];
		}
		corr += s + ' ';
	}
private:
	string corr;
};


void parse_possess()
{
	// Find possessive in sent for keyword recententities["found"].
	// Store it in recententities["owner"].

	sent_rev_it it = find(sent.rbegin(), sent.rend(), recententities["found"]) + 1;

	string owner = "your";
	if (it != sent.rend()) {
		string prev = *it;
		if (wordtypes.isType(prev, "possessivepronoun") ||
			prev[prev.size() - 1] == 's') owner = prev;
	}
	recententities["owner"] = owner;
}

sent_it parse_subjsearch(int dist = 1)
{
	// Search for the subject of the sentence in sent, looking for the
	// noun closest to and preceding recententities["found"] by at least
	// dist words. Set recententities["subj"] to the subject noun.
	// Return the iterator that points to the subject or sent.end().

	sent_rev_it rit = find(sent.rbegin(), sent.rend(), recententities["found"]) + dist;

	while (rit < sent.rend()) {
		if (wordtypes.isNoun(*rit)) {
			recententities["subj"] = *rit;
			break;
		}
		rit++;
	}

	if (rit == sent.rend()) {
		recententities["subj"] = "you";
		return sent.end();
	}
	return rit.base() - 1;
}

string parse_getnoun(sent_it it)
{
	string obj = "";

	while (it < sent.end()) {
		// cout << "Inspecting: " << *it << endl;
		if (it + 1 == sent.end()) {
			if (wordtypes.isNoun(*it)) {
				if (!recententities["object"].size())
					recententities["object"] = *it;
				if (obj.size()) obj += ' ';
				obj += *it;
			} else {
				string last = *it;
				last[0] = toupper(last[0]);
				cout << last <<  " what?" << endl;
				string s;
				getline(cin, s);
				obj += s;
			}
		} else if (*it == "to") {
			if (obj.size()) obj += ' ';
			obj += "to";
		} else if (wordtypes.isType(*it, "preposition")) {
			// skip
		} else if (!wordtypes.isNoun(*it)) {
			string s = *it;
			if (s == "a" || s == "some") s = "this";
			if (s == "one") s = "that";
			if (obj.size()) obj += ' ';
			obj += s;
		} else {
			// Emacs Psychiatrist does not iterate further at
			// this point. We collect the complete object phrase!
			if (!recententities["object"].size())
				recententities["object"] = *it;
			if (obj.size()) obj += ' ';
			obj += *it;
		}
		it++;
	}

	if (!recententities["object"].size()) {
		recententities["object"] = "something";
	}
	return obj;
}

void doctor_remember(const string);

void parse_svo(bool remem = true)
{
	// Find subject, verb and object in vector sent, with focus on
	// recententities["found"].

	// "obj" is the complete object phrase, whereas "object" is the noun
	// of the object phrase (which is only used to define new meanings),
	// for example:
	//	Patient: Linux is related to Unix and its variants.
	// => object = "Unix", so we can set meanings["linux"] to "mach".

	recententities["obj"] = "";
	recententities["object"] = "";
	recententities["subj"] = "";
	recententities["verb"] = "";
	recententities["lover"] = "";
	recententities["feared"] = "";

	// cout << "Searching subject...";
	sent_it it = parse_subjsearch();

	if (it == sent.end()) {
		it = sent.begin();
		remem = false;
	}

	while (it < sent.end()) {
		if (wordtypes.isType(*it, "verb")) break;
		it++;
	}

	if (it < sent.end()) {
		recententities["verb"] = *it;
		recententities["obj"] = parse_getnoun(++it);
	}

	if (recententities["object"] == "i") {
		recententities["object"] == "me";
	} else if (recententities["subj"] == "me") {
		recententities["subj"] == "i";
	}
	if (remem) {
		doctor_remember(recententities["subj"] + ' ' + recententities["verb"] + ' ' + recententities["obj"]);
	}
}

void doctor_remember(const string s)
{
	synonyms["history"].add(s);
}

void doctor_forget()
{
	synonyms["history"].removeLast();
}


void exec_branch(const string &);

/*
	The branch_* procedures are called as "reactions" to the first word
	of the input (chosen randomly) that has an associated meaning.
*/


void branch_alcohol()
{
	printline("$drnk");
}

void branch_chat()
{
	printline("$chatlist");
}

void branch_short();
void branch_conj()
{
	sent_it it = find(sent.begin(), sent.end(), recententities["found"]);
	if (++it > (sent.end() - 3)) {
		branch_short();
		return;
	}

	vector<string> newsent(it, sent.end());
	sent = newsent;
	if (sent[0] == "of") {
		printline("are you $sure that is the real reason ?");
		string newthing;
		it = sent.begin() + 1;
		while (it != sent.end()) {
			if (newthing.size()) newthing += ' ';
			newthing += *it;
		}
		synonyms["things"].add(newthing);
	} else {
		printline("$beclist");
	}
}

void branch_death()
{
	static bool suicideflag = false;
	string found = recententities["found"];
	if (!suicideflag && ((found == "suicide") ||
	   ((found == "kill" || found == "killing") &&
	    sentcontains("yourself")))) {
		suicideflag = true;
		printline("If you are really suicidal, you might want to contact the Samaritans via E-mail: jo@samaritans.org or, at you option, via anonymous E-Mail: samaritans@anon.twwells.com . Or find a Befrienders crisis center at http://www.befrienders.org/ . $please $continue .");
	} else {
		printline("$deathlist");
	}
}

void branch_desire()
{
	sent_it it = find(sent.begin(), sent.end(), recententities["found"]);
	if (++it >= sent.end()) {
		exec_branch(synonyms["whereoutp"].current());
		return;
	}

	if (*it == "a" || *it == "an") {
		it = sent.insert(it, "have");
		sent.insert(it, "to");
		parse_svo(false);
		doctor_remember(recententities["subj"] + " would like " + recententities["obj"]);
		printline("$whywant");
	} else if (*it != "to") {
		exec_branch(synonyms["whereoutp"].current());
		return;
	} else {
		parse_svo(false);
		doctor_remember(recententities["subj"] + " would like " + recententities["obj"]);
		printline("$whywant");
	}
}

void branch_drug()
{
	printline("$drugs");
	doctor_remember("you used " + recententities["found"]);
}

void branch_eliza()
{
	static bool elizaflag = false;
	if (elizaflag) {
		printline("$elizalist");
	} else {
		printline("/found ? hah ! $please $continue .");
		elizaflag = true;
	}
}

void branch_fear()
{
	string feared = "something";
	sent_it it = find(sent.begin(), sent.end(), recententities["found"]) + 1;

	if (it != sent.end()) {
		if (wordtypes.isType(*it, "preposition")) {
			feared = parse_getnoun(++it);
		} else if (wordtypes.isType(*it, "article")) {
			feared = *it + ' ' + parse_getnoun(++it);
		}
	}

	recententities["feared"] = feared;
	printline("$fears");
	doctor_remember("you were afraid of " + recententities["feared"]);
}

void branch_family()
{
	parse_possess();
	printline("$famlist");
}

void branch_foul()
{
	printline("$foullist");
}

void branch_huh();
void branch_hate()
{
	parse_svo();
	if (sentcontains("not")) {
		doctor_forget();
		branch_huh();
	} else if (recententities["subj"] == "you") {
		printline("why do you /verb /obj ?");
	} else {
		printline("$whysay /subj /verb /obj ?");
	}
}

void doctor_hates1()
{
	printline("$whysay /subj /verb /obj ?");
}

void branch_hates()
{
	parse_svo();
	doctor_hates1();
}

void branch_howdy()
{
	static bool howdyflag = false;
	if (!howdyflag) {
		printline("$hello what brings you to see me ?");
		howdyflag = true;
	} else {
		printline("$ibelieve we've introduced ourselves already .");
		printline("$please $describe $things .");
	}
}

void branch_huh()
{
	if (sent.size() < 9) {
		printline("$huhlist");
	} else {
		printline("$longhuhlist");
	}
}

void doctor_qloves()
{
	printline("$bother /subj /verb /obj ?");
}

void branch_love()
{
	parse_svo();

	if (sentcontains("not")) {
		doctor_forget();
		branch_huh();
		return;
	} else if (sentcontains("to")) {
		doctor_hates1();
		return;
	}

	string lover;
	if (recententities["obj"] == "something") {
		recententities["obj"] = "this person you love";
	} else if (recententities["subj"] == "you") {
		recententities["lover"] = recententities["obj"];
		if (recententities["lover"] == "this person you love") {
			recententities["lover"] = "your partner";
			doctor_forget();
			printline("with whom are you in love ?");
		} else {
			printline("$please $describe $relation /lover .");
		}
	} else if (recententities["subj"] == "i") {
		printline("we were discussing you !");
	} else {
		doctor_forget();
		recententities["obj"] = "someone";
		recententities["verb"] += 's';
		doctor_qloves();
	}
}

void branch_loves()
{
	parse_svo();
	doctor_qloves();
}

string plural_of(const string &word)
{
	char lastletter = word[word.size() - 1];
	if (lastletter == 's') {
		if (word[word.size() - 2] == 's') {
			return word + "es";
		} else {
			return word;
		}
	} else if (lastletter == 'y') {
		return word.substr(0, word.size() - 1) + "ies";
	} else if (lastletter == 'x') {
		return word + "es";
	} else {
		return word + 's';
	}
}

void branch_mach()
{
	recententities["found"] = plural_of(recententities["found"]);
	printline("$machlist");
}

void branch_math()
{
	printline("$mathlist");
}

void branch_mood()
{
	printline("$moods");
	doctor_remember("you felt " + recententities["found"]);
}

void branch_remem()
{
	if (synonyms["history"].size() == 0) {
		branch_huh();
	} else {
		printline("$remlist");
	}
}

void branch_rms()
{
	static bool rmsflag = false;
	if (rmsflag) {
		printline("$stallmanlist");
	} else {
		rmsflag = true;
		printline("do you know Stallman ?");
	}
}

void branch_rthing()
{
	printline("$thlist");
}

void branch_school()
{
	printline("$schoollist");
}

void branch_sexverb()
{
	printline("$sexlist");
}

void branch_sexnoun()
{
	branch_sexverb();
}

void branch_short()
{
	static bool mad = false;
	static unsigned int lastshortline = 0;
	static unsigned int numshortlines = 0;

	if (lastshortline == (linecount - 1)) {
		numshortlines++;
	} else {
		numshortlines = 1;
	}
	lastshortline = linecount;

	if (numshortlines > 6) {
		if (!mad) {
			printline("$areyou just trying to see what kind of things i have in my vocabulary ? ");
			printline("please try to carry on a reasonable conversation !");
			mad = true;
		} else {
			printline("i give up . you need a lesson creative writing ...");
		}
		return;
	}


	if (!sent.size()) {
		printline("$shortlist");
		return;
	}

	if (sent[0] == "yes") {
		printline("$isee $inter $whysay this is so ?");
	} else if (sent[0] == "because") {
		printline("$shortbeclist");
	} else if (sent[0] == "no") {
		printline("$neglist");
	} else if (meanings.find(sent[0]) != meanings.end() && meanings[sent[0]] == "howdy") {
		branch_howdy();
	} else {
		printline("$shortlist");
	}
}

void branch_sports()
{
	printline("$sportslist");
}

void branch_state()
{
	printline("$states");
	doctor_remember("you were " + recententities["found"]);
}

void branch_symptoms()
{
	printline("$maybe you should consult a doctor of medicine, i am a psychiatrist .");
}

void branch_toke()
{
	printline("$toklist");
}

void branch_when()
{
	sent_it it = find(sent.begin(), sent.end(), recententities["found"]);

	if (++it > sent.end() - 2) {
		branch_short();
	} else {
		vector<string> newsent(it, sent.end());
		sent = newsent;
		printline("$whatwhen /sent ?");
	}
}

void doctor_zippy()
{
	static bool zippyflag = false;
	if (zippyflag) {
		printline("$zippylist");
	} else {
		zippyflag = true;
		printline("yow ! are we interactive yet ?");
	}
}



typedef void (* branchpointer)(void);
map<string, branchpointer> branches;

void init_branches()
{
	branches["family"] = branch_family; branches["foul"] = branch_foul;
	branches["huh"] = branch_huh; branches["rthing"] = branch_rthing;
	branches["remem"] = branch_remem; branches["howdy"] = branch_howdy;
	branches["when"] = branch_when; branches["conj"] = branch_conj;
	branches["alcohol"] = branch_alcohol; branches["rms"] = branch_rms;
	branches["desire"] = branch_desire; branches["remem"] = branch_remem;
	branches["drug"] = branch_drug; branches["toke"] = branch_toke;
	branches["state"] = branch_state; branches["mood"] = branch_mood;
	branches["fear"] = branch_fear; branches["hate"] = branch_hate;
	branches["love"] = branch_love; branches["symptoms"] = branch_symptoms;
	branches["eliza"] = branch_eliza; branches["school"] = branch_school;
	branches["sports"] = branch_sports; branches["chat"] = branch_chat;
	branches["math"] = branch_math; branches["sexverb"] = branch_sexverb;
	branches["sexnoun"] = branch_sexnoun; branches["death"] = branch_death;
	branches["mach"] = branch_mach; branches["loves"] = branch_loves;
}

// This imitates the central "funcall" instruction of Emacs doctor.
// It calls a procedure based on a found keyword.

void exec_branch(const string &s)
{
	if (branches.find(s) != branches.end()) {
		branches[s]();
	} else {
		cout << "WARNING: Branch '" << s << "' is not defined!" << endl;
	}
}

bool isvowel(const char &c)
{
	return (c == 'a' || c == 'e' || c == 'i' || c == 'o' || c == 'u');
}

void fix_input()
{
	// Remove some interjections and change some words

	sent_it it = sent.begin();
	if (*it == "me") {
		*it = "i";
	} else if (*it == "him") {
		*it = "he";
	} else if (*it == "her") {
		*it = "she";
	} else if (*it == "them") {
		*it = "they";
	}

	string inter[] = { "okay", "well", "sigh", "hmm", "hmmm", "hmmmm", "hmmmmm", "gee", "sure", "great", "oh", "fine", "ok", "yes", "no", "sorry", "" };
	unsigned int i = 0;
	while (inter[i] != "") {
		if (sent.front() == inter[i++]) sent.erase(sent.begin());
		if (!sent.size()) break;
	}
}

void fix_input_2()
{
	// Swap perspective

	// This test is necessary because of signed/unsigned issues in
	// (sent.size() - 1)!
	if (!sent.size()) return;

	// Only loop to the last but one word because the next word is affected.
	for (unsigned int i = 0; i < (sent.size() - 1); i++) {
		if (sent[i] == "me" && wordtypes.isType(sent[i + 1], "verb")) {
			sent[i] = "i";
			continue;
		}
		if (sent[i] == "you") {
			string s = sent[i + 1];
			if (s == "am" || s == "be" || s == "been" ||
			    s == "is")	sent[i + 1] = "are";
			if (s == "has") sent[i + 1] = "have";
			if (s == "was") sent[i + 1] = "were";
		} else if (sent[i] == "i") {
			string s = sent[i + 1];
			if (s == "are" || s == "is" || s == "be" ||
			    s == "been") sent[i + 1] = "am";
			if (s == "were") sent[i + 1] = "was";
			if (s == "has")  sent[i + 1] = "have";
		} else if (wordtypes.isType(sent[i], "verb") &&
			(sent[i + 1] == "i") && (i + 2 < sent.size()) &&
			(!wordtypes.isType(sent[i + 2], "verb"))) {
				sent[i + 1] = "me";
		} else if (sent[i] == "a" && isvowel((sent[i + 1])[0])) {
			sent[i] = "an";
		} else if (sent[i] == "an" && !isvowel((sent[i + 1])[0])) {
			sent[i] = "a";
		}
	}
}


void doctor_shorten()
{
	// Find a suitable position to split the sentence
	string conj[] = {"because", "but", "however", "besides", "anyway", "until", "while", "that", "except", "why", "how", "" };
	unsigned int i = 0;
	while (conj[i] != "") {
		sent_it it = find(sent.begin(), sent.end(), conj[i]);
		if (it != sent.end()) {
			vector<string> testsent;
			testsent.assign(it, sent.end());
			if (testsent.size() > 3) {
				sent = testsent;
				return;
			}
		}
		i++;
	}
}

bool isdefine()
{
	string defwords[] = {"means", "applies", "mean", "refers", "refer",
		"related", "similar", "defined", "associated", "linked",
		"like", "same", "" };

	unsigned int i = 0;
	while (defwords[i] != "") {
		if (sentcontains(defwords[i++])) {
			recententities["found"] = defwords[--i];
			return true;
		}
	}
	return false;
}

void do_define()
{
	parse_svo(false);
	string subj = recententities["subj"];
	string object = recententities["object"];
	if (wordtypes.isNoun(subj) && !(wordtypes.isType(subj, "pronoun")) &&
	subj.size() && meanings.find(object) != meanings.end()) {
		meanings[subj] = meanings[object];
	}
}

int do_reply(string input)
{
	if (input == "foo") {
		printline("Bar! $please $continue .");
	}

	input = trim(input);
	bool wasquestion = false;
	if (input.size() && input[input.size() - 1] == '?') wasquestion = true;

	for (string::iterator it = input.begin(); it != input.end(); it++) {
		// eliminate all non-alpha non-digit chars except apastrophes
		if (ispunct(*it) && (*it != '\'') && (*it != '`')) *it = ' ';
		if (isalpha(*it)) *it = tolower(*it);
	}
	// trim again (to remove punctuation that is now blank)
	input = trim(input);
	// cout << input << endl;

	if (input != "parse") {
		recententities["bak"] = input;
	} else {
		cout << "subj: " << recententities["subj"] << ", verb: " << recententities["verb"] << endl;
		cout << "object phrase: " << recententities["obj"] << ", noun form: " << recententities["object"] << endl;
		cout << "current keyword is: " << recententities["found"] << ", most recent possessive is: " << recententities["owner"] << endl;
		cout << "sentence used was: " << recententities["bak"] << endl;
		return 1;
	}

	if (synonyms["howareyoulist"].isMember(input)) {
		printline("i'm ok . $describe yourself .");
		return 1;
	}

	if (synonyms["whocreator"].isMember(input)) {
		printline("I was a part of Emacs. Markus Triska converted me to C++.");
		printline("You can send him a message via triska@gmx.at");
		return 1;
	}

	if (synonyms["byelist"].isMember(input)) {
		printline("$bye");
		return 0;
	}

	sent.clear();
	for_each_word(input, push_to_sent);

	if (sent.size() && sent[0] == "yow") {
		doctor_zippy();
		return 1;
	}

	if (sent.size() < 2) {
		branch_short();
		return 1;
	}

	if (sent.size()) fix_input();

	if (isdefine()) do_define();

	if (sent.size() > 12) doctor_shorten();

	string first = sent[0], second = sent[1];

	if (synonyms["byelist"].isMember(first)) {
		printline("$bye");
		return 0;
	}

	if ((first == "you") && (synonyms["abusewords"].isMember(second))) {
		recententities["found"] = second;
		printline("$abuselist");
		return 1;
	}

	if (first == "whatmeans") {
		string meaning = "nothing";
		if (meanings.find(second) != meanings.end()) {
			meaning = meanings[second];
		}
		printline("the word " + second + " means " + meaning + " to me .");
		return 1;
	}

	if ((first == "do") && (second == "not")) {
		if (!(rand() % 3)) {
			printline("are you $afraidof that ?");
		} else if (!(rand() % 2)) {
			printline("don't tell me what to do . i am the psychiatrist here!");
			printline("$thlist");
		} else {
			string what = (sent.size() >= 3) ? sent[2] : string("");
			string msg = (string("$whysay that i shouldn't ") + what  + '?');
			printline(msg);
		}
		return 1;
	}

	if (wasquestion || wordtypes.isType(first, "questionwords")) {
		printline("$qlist");
		return 1;
	}


	WordCorrector corr;
	corr = for_each_word(input, corr);
	input = corr.corrected();
	// cout << "Corrected input: " << input << endl;

	sent.clear();
	for_each_word(input, push_to_sent);

	if (!sentcontains("me") && (!sentcontains("i")) && sentcontains("am")) {
		replace(sent.begin(), sent.end(), string("am"), string("are"));
	}

	if (sentcontains("am")) {
		replace(sent.begin(), sent.end(), string("me"), string("i"));
	}


	vector<string> getmeaning(sent);
	random_shuffle(getmeaning.begin(), getmeaning.end());

	for (sent_it it = getmeaning.begin(); it != getmeaning.end(); it++) {
		if (meanings.find(*it) != meanings.end()) {
			recententities["found"] = *it;
			// cout << "Found " << *it << ", meaning: " << meanings[*it] << endl;
			exec_branch(meanings[*it]);
			return 1;
		}
	}

	exec_branch(synonyms["whereoutp"].current());

	return 1;
}

int main()
{
	init_branches();
	init_corrections();
	init_meanings();
	init_synonyms();
	init_wordtypes();

	cout << "I'm Freud " << VERSION << ". Please describe your problem!" << endl;
	while (true) {
		string input;
		getline(cin, input);
		linecount++;
		if (!do_reply(input)) break;
	}
	return 0;
}
